# -*- coding: utf-8 -*-
from ..utils import get_file_type, get_option, detect_encoding, detect_delimiter
import zipfile
import csv
import json
import orjson
import jsonlines
import bson
import logging

DEFAULT_ENCODING = 'utf8'
DEFAULT_DELIMITER = ','

class IterableData:
    """Iterable data reader (CSV/JSON lines, BSON"""
    def __init__(self, filename, options={}, autodetect=True, autodetect_limit=100000):
        """Creates iterable object from CSV, JSON lines or BSON file.
        """
        self.autodetect = autodetect
        self.autodetect_limit = autodetect_limit
        self.options = options
        self.init(filename, options)
        pass


    def init(self, filename, options):
        f_type = get_file_type(filename) if options['format_in'] is None else options['format_in']
        self.filetype = f_type
        if options['zipfile']:
            z = zipfile.ZipFile(filename, mode='r')
            fnames = z.namelist()
            if f_type == 'bson':
                self.fileobj = z.open(fnames[0], 'rb')
            else:
                # FIXME
                self.fileobj = z.open(fnames[0], 'r')
        else:
            if f_type == 'bson':
                self.fileobj = open(filename, 'rb')
            else:
                if 'encoding' in options.keys() and options['encoding']:
                    encoding = get_option(options, 'encoding')
                else:
                    if self.autodetect:
                        detected_enc = detect_encoding(filename, limit=self.autodetect_limit)
                        encoding = detected_enc['encoding'] if detected_enc else DEFAULT_ENCODING
                        if f_type == 'csv':
                            detected_del = detect_delimiter(filename, encoding)
                            delimiter = detected_del if detected_del else DEFAULT_DELIMITER
                            self.delimiter = delimiter
                    else:
                        encoding = DEFAULT_ENCODING
                        if f_type == 'csv':
                            delimiter = DEFAULT_DELIMITER
                    logging.debug('Detected encoding %s' % (detected_enc['encoding']))
                self.encoding = encoding
                self.fileobj = open(filename, 'r', encoding=encoding)

    def iter(self):
        if self.filetype == 'csv':
            reader = csv.DictReader(self.fileobj, delimiter=self.delimiter)
            return iter(reader)
        elif self.filetype == 'jsonl':
            return jsonlines.Reader(self.fileobj)
        elif self.filetype == 'bson':
            return bson.decode_file_iter(self.fileobj)

    def close(self):
        """Closes file object"""
        self.fileobj.close()



class BSONWriter:
    def __init__(self, fileobj):
        self.fo = fileobj

    def write(self, item):
        rec = bson.BSON.encode(j)
        self.fo.write(rec)


class DataWriter:
    """Data writer (CSV/JSON lines, BSON"""
    def __init__(self, fileobj, filetype, delimiter=',', fieldnames=None):
        """Creates iterable object from CSV, JSON lines or BSON file.
        """
        self.filetype = filetype
        self.fieldnames = fieldnames
        self.fileobj = fileobj
        if self.filetype == 'csv':
            self.writer = csv.DictWriter(self.fileobj, delimiter=delimiter, fieldnames=fieldnames)
        elif self.filetype == 'jsonl':
            self.writer = jsonlines.Writer(self.fileobj)
        elif self.filetype == 'bson':
            self.writer = BSONWriter(self.fileobj)
        pass


#    def write_item(self, row):

    def write_items(self, outdata):
        if len(outdata) == 0:
            return
        if self.filetype == 'csv':
            self.writer.writeheader()
            if isinstance(outdata[0], str):
                for rawitem in outdata:
                    item = {self.fieldsnames[0]: rawitem}
                    self.writer.writerow(item)
            elif type(outdata[0]) == type([]):
                for rawitem in outdata:
                    item = dict(zip(self.fieldsnames, rawitem))
                    self.writer.writerow(item)
            else:
                self.writer.writerows(outdata)
        elif self.filetype in ['jsonl', 'bson']:
            # If our data is just array of strings, we just transform it to dict
            if isinstance(outdata[0], str):
                for rawitem in outdata:
                    item = {self.fieldnames[0]: rawitem}
                    self.writer.write(item)
#                    handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
            elif type(outdata[0]) == type([]):
                for rawitem in outdata:
                    item = dict(zip(self.fieldnames, rawitem))
                    self.writer.write(item)
#                    handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
            else:
                for item in outdata:
                    self.writer.write(item)
#                    handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))


if __name__ == "__main__":
    f = open('outtest.jsonl', 'w')
    writer = DataWriter(f, filetype='jsonl', fieldnames=['name', 'value'])
    writer.write_items([{'name' : 'Cat', 'value' : 15}])
    pass



from xlrd import open_workbook
from operator import itemgetter, attrgetter
import csv
import zipfile
import sys
import orjson
import bson
import logging
#from xmlr import xmliter
import xml.etree.ElementTree as etree
from collections import defaultdict
from ..utils import get_file_type, get_option, write_items, get_dict_value, strip_dict_fields, dict_generator
import dictquery as dq
from runpy import run_path

class Transformer:
    def __init__(self):
        pass


    def script(self, fromfile, options={}):
        """Run certain script against selected file"""
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        if options['zipfile']:
            z = zipfile.ZipFile(fromfile, mode='r')
            fnames = z.namelist()
            if f_type == 'bson':
                infile = z.open(fnames[0], 'rb')
            else:
                infile = z.open(fnames[0], 'r')
        else:
            if f_type == 'bson':
                infile = open(fromfile, 'rb')
            else:
                infile = open(fromfile, 'r', encoding=get_option(options, 'encoding'))
        to_file = get_option(options, 'output')
        if to_file:
            to_type = get_file_type(to_file)
            if not to_file:
                print('Output file type not supported')
                return
            if to_type == 'bson':
                out = open(to_file, 'wb')
            elif to_type == 'jsonl':
                out = open(to_file, 'wb')
            else:
                out = open(to_file, 'w', encoding='utf8')
        else:
            to_type = f_type
            out = sys.stdout
 #       fields = options['fields'].split(',') if options['fields'] else None
        script = run_path(options['script'])
        __process_func = script['process']
        delimiter = get_option(options, 'delimiter')
        if f_type == 'csv':
            reader = csv.DictReader(infile, delimiter=delimiter)
            if to_type == 'csv':
                writer = csv.DictWriter(out, fieldnames=fields, delimiter=delimiter)
                writer.writeheader()
            n = 0
            for r in reader:
                n += 1
                if n % 10000 == 0:
                    logging.info('apply script: processing %d records of %s' % (n, fromfile))
                item = __process_func(r)
                if to_type == 'csv':
                    writer.writerow(item)
                elif to_type == 'jsonl':
                    out.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE))
        elif f_type == 'jsonl':
            n = 0
            for l in infile:
                n += 1
                if n % 10000 == 0:
                    logging.info('apply script: processing %d records of %s' % (n, fromfile))
                r = orjson.loads(l)
                item = __process_func(r)
                out.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE))
        elif f_type == 'bson':
            bson_iter = bson.decode_file_iter(infile)
            n = 0
            for r in bson_iter:
                n += 1
                if n % 10000 == 0:
                    logging.info('apply script: processing %d records of %s' % (n, fromfile))
                item = __process_func(r)
                out.write(str(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE)))
        else:
            logging.info('File type not supported')
            return
        logging.debug('select: %d records processed' % (n))
        out.close()





#!/usr/bin/env python
# -*- coding: utf8 -*-
import click
import logging

from .cmds.converter import Converter
from .cmds.selector import Selector
from .cmds.transformer import Transformer
from .cmds.analyzer import Analyzer
from .cmds.textproc import TextProcessor
from .cmds.validator import Validator
from .cmds.schemer import Schemer

#logging.getLogger().addHandler(logging.StreamHandler())
logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    level=logging.DEBUG)

def enableVerbose():
    logging.basicConfig(
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        level=logging.DEBUG)


@click.group()
def cli1():
    pass

@cli1.command()
@click.argument('input')
@click.argument('output')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV, default ','")
@click.option('--encoding', '-e', default='utf8', help="Input and output encoding, default is utf8")
@click.option('--verbose', '-v', count=False, help='Verbose output. Print additional info')
@click.option('--prefix-strip', default=True, is_flag=True, help="Strip prefix from XML files")
@click.option('--fields', '-f', default=None, help="Fieldnames from XLS conversion")
@click.option('--start-line', default=0, help="Start line, used for XLS and XLSX conversion")
@click.option('--skip-end-rows', default=0, help="Skip rows at the end of xls file")
@click.option('--start-page', default=0, help="Start page, used for XLS and XLSX conversion")
@click.option('--tagname',  default=None, help="Object tagname, needed for xml2json")
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('--format-out',  default=None, help="Format of output file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=False, help="Used to say input file is .zip file and that data file is inside")
def convert(input, output, delimiter, encoding, verbose, prefix_strip, fields, start_line, skip_end_rows, start_page, tagname, format_in, format_out, zipfile):
    """Converts one file to another. Supports XML, CSV, JSON, BSON"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['encoding'] = encoding
    options['prefix_strip'] = prefix_strip
    options['start_line'] = start_line
    options['skip_end_rows'] = skip_end_rows
    options['start_page'] = start_page
    options['tagname'] = tagname
    options['fields'] = fields
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    acmd = Converter()
    acmd.convert(input, output, options)
    pass


@click.group()
def cli2():
    pass

@cli2.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--fields', '-f', default=None, help="Fieldnames, delimiter by ','")
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default=None, help="Input and output encoding")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('--format-out',  default=None, help="Format of output file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('--filter',  default=None, help="Filter input file with dict query")
def uniq(input, output, fields, delimiter, encoding, verbose, format_in, format_out, zipfile, filter):
    """Returns all unique files of certain field(s)"""
    if verbose:
        enableVerbose()
    options = {}
    options['output'] = output
    options['fields'] = fields
    options['delimiter'] = delimiter
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['filter'] = filter
    acmd = Selector()
    acmd.uniq(input, options)
    pass

@click.group()
def cli3():
    pass

@cli3.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=None, help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default=None, help="Input and output encoding")
@click.option('--limit', '-l', default=1000, help="Limit of lines used to detect headers of JSON/BSON files only")
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('--format-out',  default=None, help="Format of output file, if set, replaces autodetect")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
def headers(input, output, delimiter, encoding, limit, format_in, format_out, verbose, zipfile):
    """Returns fieldnames of the file. Supports XML, CSV, JSON, BSON"""
    if verbose:
        enableVerbose()
    options = {}
    options['output'] = output
    options['delimiter'] = delimiter
    options['encoding'] = encoding
    options['limit'] = limit
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    acmd = Selector()
    acmd.headers(input, options)
    pass

@click.group()
def cli4():
    pass

@cli4.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--dictshare', '-s', 'dictshare', default=None, help="Uniqness level of certain field to detect that it's dict")
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('--format-out',  default=None, help="Format of output file, if set, replaces autodetect")
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('--checkdates', is_flag=True, help="Significantly slow down process, buy identifies dates fields from text. Not used by default")
def stats(input, output, dictshare, format_in, format_out, delimiter, verbose, zipfile, checkdates):
    """Returns detailed stats on selected dataset"""
    if verbose:
        enableVerbose()
    options = {}
    options['output'] = output
    options['dictshare'] = dictshare
    options['zipfile'] = zipfile
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['delimiter'] = delimiter
    options['checkdates'] = checkdates
    options['verbose'] = verbose
    acmd = Analyzer(nodates=not checkdates)
    acmd.analyze(input, options)
    pass


@click.group()
def cli5():
    pass

@cli5.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default='utf8', help="Input and output encoding")
@click.option('--filter',  default=None, help="Filter input file with dict query")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
def flatten(input, output, delimiter, encoding, filter, verbose):
    """Flatten data records. Write them as one value per row"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['output'] = output
    options['encoding'] = encoding
    options['filter'] = filter
    acmd = TextProcessor()
    acmd.flatten(input, options)
    pass


@click.group()
def cli6():
    pass

@cli6.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default=None, help="Input and output encoding")
@click.option('--fields', '-f', default=None, help="Fieldnames, delimiter by ','")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('--format-out',  default=None, help="Format of output file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('--filter',  default=None, help="Filter input file with dict query")
def frequency(input, output, fields, delimiter, encoding, verbose, format_in, format_out, zipfile, filter):
    """Field value frequency calc"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['filter'] = filter
    acmd = Selector()
    acmd.frequency(input, options)
    pass


@click.group()
def cli7():
    pass

@cli7.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=None, help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default=None, help="Input and output encoding")
@click.option('--fields', '-f', default=None, help="Fieldnames, delimiter by ','")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('--format-out',  default=None, help="Format of output file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('--filter',  default=None, help="Filter input file with dict query")
def select(input, output, fields, delimiter, encoding, verbose, format_in, format_out, zipfile, filter):
    """Select or re-order columns from file. Supports CSV, JSONl, BSON"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['filter'] = filter
    acmd = Selector()
    acmd.select(input, options)
    pass


@click.group()
def cli8():
    pass

@cli8.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default='utf8', help="Input and output encoding")
@click.option('--fields', '-f', default=None, help="Fieldnames, delimiter by ','")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('--gzip', '-g', 'gzipfile', is_flag=True, help="Used to say input file is gzipped file")
@click.option('-c', '--chunksize', default=10000, help='Default chunk size of file to split. Used if field value not given')
@click.option('--filter',  default=None, help="Filter input file with dict query")
def split(input, output, fields, delimiter, encoding, verbose, format_in, zipfile, gzipfile, chunksize, filter):
    """Splits the given file with data into chunks."""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['gzipfile'] = gzipfile
    options['chunksize'] = chunksize
    options['filter'] = filter
    acmd = Selector()
    acmd.split(input, options)
    pass

@click.group()
def cli9():
    pass


@cli9.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default='utf8', help="Input and output encoding")
@click.option('--fields', '-f', default=None, help="Fieldnames, delimiter by ','")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('-r', '--rule',  default=None, required=True, help="Validation rule")
@click.option('--filter',  default=None, help="Filter input file with dict query")
@click.option('-m', '--mode',  default='invalid', help="Mode of validation output: invalid, all, or stats")
def validate(input, output, fields, delimiter, encoding, verbose, format_in, zipfile, rule, filter, mode):
    """Validates selected field against validation rule"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['filter'] = filter
    options['rule'] = rule
    options['mode'] = mode
    acmd = Validator()
    acmd.validate(input, options)
    pass

@click.group()
def cli10():
    pass


@cli10.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default='utf8', help="Input and output encoding")
@click.option('--fields', '-f', default=None, help="Fieldnames, delimiter by ','")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('-s', '--script',  default=None, required=True, help="Script to run")
@click.option('--filter',  default=None, help="Filter input file with dict query")
def apply(input, output, fields, delimiter, encoding, verbose, format_in, zipfile, script, filter):
    """Runs script against each record of input file"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['filter'] = filter
    options['script'] = script
    acmd = Transformer()
    acmd.script(input, options)
    pass


@click.group()
def cli11():
    pass

@cli11.command()
@click.argument('input')
@click.option('--output', '-o', 'output', default=None, help='Output to this file')
@click.option('--delimiter', '-d', default=',', help="CSV delimiter if convert from CSV")
@click.option('--encoding', '-e', default='utf8', help="Input and output encoding")
@click.option('--verbose', '-v', count=True, help='Verbose output. Print additional info on command execution')
@click.option('--format-in',  default=None, help="Format of input file, if set, replaces autodetect")
@click.option('-z', '--zipfile', 'zipfile', is_flag=True, help="Used to say input file is .zip file and that data file is inside")
@click.option('--stype',  default='cerberus', help="Type of the schema: cerberus")
def scheme(input, output, delimiter, encoding, verbose, format_in, zipfile, stype):
    """Runs script against each record of input file"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['stype'] = stype
    acmd = Schemer()
    acmd.generate_scheme(input, options)
    pass


cli = click.CommandCollection(sources=[cli1, cli2, cli3, cli4, cli5, cli6, cli7, cli8, cli9, cli10, cli11])

#if __name__ == '__main__':
#    cli()




from ..utils import get_file_type, get_option
from ..constants import DATE_PATTERNS, DEFAULT_DICT_SHARE
from datetime import datetime
import orjson
import csv
import bson

def get_keys(adict, prefix=None):
    keys = {}
    for k, v in adict.items():
        fullk = '.'.join([prefix, k]) if prefix else k
        keys[fullk] = 1
        if type(v) == type({}):
            for ak in get_keys(v, fullk):
                keys[ak] = 1
        elif type(v) == type([]):
            for item in v:
                if type(item) == type({}):
                    for ak in get_keys(item, fullk):
                        keys[ak] = 1
        else:
            print((u'%s\t%s' % (fullk, str(v))))
    return keys


class TextProcessor:
    def __init__(self):
        pass

    def flatten(self, filename, options):
        if filename[-5:] == '.bson':
            f = open(filename, 'rb')
        else:
            f = open(filename, 'r', encoding=get_option(options, 'encoding'))
        i = 0
        if filename[-5:] == 'jsonl':
            for r in f:
                allkeys = {}
                i += 1
                rec = orjson.loads(r)
                for k in get_keys(rec):
                    v = allkeys.get(k, 0)
                    allkeys[k] = v + 1
                for k, v in allkeys.items():
                    print('\t'.join([k, str(v)]))
        elif filename[-5:] == '.bson':
            for rec in bson.decode_file_iter(f):
                allkeys = {}
                for k in get_keys(rec):
                    v = allkeys.get(k, 0)
                    allkeys[k] = v + 1
                for k, v in allkeys.items():
                    print('\t'.join([k, str(v)]))
        elif filename[-5:] == '.json':
            allkeys = {}
            rec = orjson.loads(f.read())
            for k in get_keys(rec):
                v = allkeys.get(k, 0)
                allkeys[k] = v + 1
            for k, v in allkeys.items():
                print('\t'.join([k, str(v)]))
        elif filename[-4:] == '.csv':
            dr = csv.DictReader(f, get_option(options, 'delimiter'))
            keys = dr.fieldnames
            for r in dr:
                for key in keys:
                    print((u'%s\t%s' % (key, r[key])))




import csv
import json
import orjson
import logging
import jsonlines
import pandas
# from xmlr import xmliter
import xml.etree.ElementTree as etree
from collections import defaultdict

import bson
from xlrd import open_workbook as load_xls

PREFIX_STRIP = True
PREFIX = ""
from ..utils import get_file_type

LINEEND = '\n'.encode('utf8')

def __copy_options(user_options, default_options):
    """If user provided option so we use it, if not, default option value should be used"""
    for k in default_options.keys():
        if k not in user_options.keys():
            user_options[k] = default_options[k]
    return user_options


def etree_to_dict(t, prefix_strip=True):
    tag = t.tag if not prefix_strip else t.tag.rsplit('}', 1)[-1]
    d = {tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            #            print(dir(dc))
            for k, v in dc.items():
                if prefix_strip:
                    k = k.rsplit('}', 1)[-1]
                dd[k].append(v)
        d = {tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[tag].update(('@' + k.rsplit('}', 1)[-1], v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            tag = tag.rsplit('}', 1)[-1]
            if text:
                d[tag]['#text'] = text
        else:
            d[tag] = text
    return d


def xml_to_jsonl(fromname, toname, options={}, default_options={'prefix_strip': True}):
    options = __copy_options(options, default_options)
    ins = open(fromname, 'rb')  # , encoding='utf-8')
    outf = open(toname, 'wb')
    n = 0
    for event, elem in etree.iterparse(ins):
        shorttag = elem.tag.rsplit('}', 1)[-1]
        if shorttag == options['tagname']:
            n += 1
            if options['prefix_strip']:
                j = etree_to_dict(elem, prefix_strip=options['prefix_strip'])
            else:
                j = etree_to_dict(elem)
            outf.write(orjson.dumps(j[shorttag]))
            outf.write(LINEEND)
        if n % 500 == 0:
            logging.info('xml2jsonl: processed %d xml tags' % (n))
    logging.info('xml2jsonl: processed %d xml tags finally' % (n))
    ins.close()
    outf.close()


def xls_to_csv(fromname, toname, options={},
               default_options={'start_line': 0, 'skip_end_rows': 0, 'delimiter': ',', 'encoding': 'utf8'}):
    options = __copy_options(options, default_options)
    b = load_xls(fromname)
    s = b.sheet_by_index(0)
    bc = open(toname, 'w', encoding=options['encoding'])
    bcw = csv.writer(bc, delimiter=options['delimiter'])
    n = 0
    for row in range(options['start_line'], s.nrows - options['skip_end_rows']):
        n += 1
        this_row = []
        for col in range(s.ncols):
            v = str(s.cell_value(row, col))
            v = v.replace('\n', ' ').strip()
            #			v = v.encode('utf8') if type(v) == type(u'') else str(v)
            this_row.append(v)
        bcw.writerow(this_row)
        if n % 10000 == 0:
            logging.info('xls2csv: processed %d records' % (n))
    bc.close()


def csv_to_bson(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ','}):
    options = __copy_options(options, default_options)
    source = open(fromname, 'r', encoding=options['encoding'])
    output = open(toname, 'wb')
    reader = csv.DictReader(source, delimiter=options['delimiter'])
    n = 0
    for j in reader:
        n += 1
        rec = bson.BSON.encode(j)
        output.write(rec)
        if n % 10000 == 0:
            logging.info('csv2bson: processed %d records' % (n))
    source.close()
    output.close()


def csv_to_jsonl(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ','}):
    options = __copy_options(options, default_options)
    source = open(fromname, 'r', encoding=options['encoding'])
    output = open(toname, 'wb')
    reader = csv.DictReader(source, delimiter=options['delimiter'])
    n = 0
    for j in reader:
        n += 1
        output.write(json.dumps(j, ensure_ascii=False).encode('utf8'))
        #        output.write(orjson.dumps(j, ensure_ascii=False).encode('utf8', ))
        output.write(u'\n'.encode('utf8'))
        if n % 10000 == 0:
            logging.info('csv2jsonl: processed %d records' % (n))
    source.close()
    output.close()


def xls_to_jsonl(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0, 'fields': None}):
    options = __copy_options(options, default_options)
    source = load_xls(fromname)
    output = open(toname, 'wb')
    sheet = source.sheet_by_index(options['start_page'])
    n = 0
    fields = options['fields'].split(',') if options['fields'] is not None else None
    for rownum in range(options['start_line'], sheet.nrows):
        n += 1
        tmp = list()
        for i in range(0, sheet.ncols):
            tmp.append(sheet.row_values(rownum)[i])
        if n == 1 and fields is None:
            fields = tmp
            continue
        l = orjson.dumps(dict(zip(fields, tmp)))
        output.write(l + LINEEND)
        #        output.write(u'\n'.encode('utf8'))
        if n % 10000 == 0:
            logging.info('xls2jsonl: processed %d records' % (n))
    output.close()


def xlsx_to_jsonl(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0}):
    from openpyxl import load_workbook as load_xlsx
    options = __copy_options(options, default_options)
    source = load_xlsx(fromname)
    output = open(toname, 'wb')
    sheet = source.active
    n = 0
    for rownum in range(options['start_line'], sheet.nrows):
        n += 1
        tmp = list()
        for i in range(0, sheet.ncols):
            tmp.append(sheet.row_values(rownum)[i])
        l = orjson.dumps(dict(zip(options['fields'], tmp)))
        output.write(l)
        output.write(LINEEND)
        if n % 10000 == 0:
            logging.info('xls2jsonl: processed %d records' % (n))
    source.close
    output.close()


def xls_to_bson(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0}):
    options = __copy_options(options, default_options)
    source = load_xls(fromname)
    output = open(toname, 'wb')
    sheet = source.sheet_by_index(options['start_page'])
    n = 0
    for rownum in range(options['start_line'], sheet.nrows):
        n += 1
        tmp = list()
        for i in range(0, sheet.ncols):
            tmp.append(sheet.row_values(rownum)[i])
        output.write(bson.BSON.encode(dict(zip(options['fields'], tmp))))
        if n % 10000 == 0:
            logging.info('xls2bson: processed %d records' % (n))
    source.close
    output.close()


def _is_flat(item):
    for k, v in item.items():
        if isinstance(v, dict) or isinstance(v, tuple) or isinstance(v, list):
            return False
    return True


def express_analyze_jsonl(filename, itemlimit=100):
    f = open(filename, 'r', encoding='utf8')
    isflat = True
    n = 0
    keys = set()
    for l in f:
        n += 1
        if n > itemlimit: break
        record = orjson.loads(l)
        if isflat:
            if not _is_flat(record):
                isflat = False
        if len(keys) == 0:
            keys = set(record.keys())
        else:
            keys = keys.union(set(record.keys()))
    f.close()
    keys = list(keys)
    keys.sort()
    return {'isflat': isflat, 'keys': keys}


def jsonl_to_csv(fromname, toname, options={},
                 default_options={'force_flat': False, 'useitems': 100, 'delimiter': ','}):
    options = __copy_options(options, default_options)
    analysis = express_analyze_jsonl(fromname, itemlimit=options['useitems'])
    if not options['force_flat'] and not analysis['isflat']:
        logging.error("File %s is not flat and 'force_flat' flag not set. File not converted")
        return
    out = open(toname, 'w', encoding='utf8')
    writer = csv.writer(out, delimiter=options['delimiter'])
    writer.writerow(analysis['keys'])
    f = open(fromname, 'r', encoding='utf8')
    keys = analysis['keys']
    n = 0
    for l in f:
        n += 1
        record = orjson.loads(l)
        item = []
        for k in keys:
            if k in record.keys():
                item.append(record[k])
            else:
                item.append('')
        writer.writerow(item)
        if n % 10000 == 0:
            logging.info('jsonl2csv: processed %d records' % (n))
    out.close()
    f.close()
    pass


def bson_to_jsonl(fromname, toname, options={}, default_options={}):
    options = __copy_options(options, default_options)
    source = open(fromname, 'rb')
    output = open(toname, 'wb')
    n = 0
    for r in bson.decode_file_iter(source):
        n += 1
        output.write(orjson.dumps(r) + LINEEND)
        if n % 10000 == 0:
            logging.info('bson2jsonl: processed %d records' % (n))
    source.close()
    output.close()

def csv_to_parquet(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ','}):
    options = __copy_options(options, default_options)
    df = pandas.read_csv(fromname, delimiter=options['delimiter'], encoding=options['encoding'])
    df.to_parquet(toname)


def jsonl_to_parquet(fromname, toname, options={},
                 default_options={'force_flat': False, 'useitems': 100}):
    options = __copy_options(options, default_options)
    df = pandas.read_json(fromname, lines=True, encoding=options['encoding'])
    df.to_parquet(toname)



CONVERT_FUNC_MAP = {
    'xls2csv': xls_to_csv,
    'xls2jsonl': xls_to_jsonl,
    'xls2bson': xls_to_bson,
    'xlsx2jsonl': xlsx_to_jsonl,
    'csv2jsonl': csv_to_jsonl,
    'csv2bson': csv_to_bson,
    'xml2jsonl': xml_to_jsonl,
    'jsonl2csv': jsonl_to_csv,
    'bson2jsonl': bson_to_jsonl,
    'csv2parquet' : csv_to_parquet,
    'jsonl2parquet': jsonl_to_parquet,
}


class Converter:
    def __init__(self):
        pass

    def convert(self, fromfile, tofile, options={}):
        fromtype = options['format_in'] if options['format_in'] is not None else get_file_type(fromfile)
        totype = options['format_out'] if options['format_out'] is not None else get_file_type(tofile)
        key = '%s2%s' % (fromtype, totype)
        func = CONVERT_FUNC_MAP.get(key, None)
        if func == None:
            logging.error('Conversion between %s and %s not supported' % (fromtype, totype))
        else:
            logging.info('Convert %s from %s to %s' % (key, fromfile, tofile))
            func(fromfile, tofile, options)



import csv
import chardet
import orjson
from .constants import SUPPORTED_FILE_TYPES
from .constants import DEFAULT_OPTIONS

def detect_encoding(filename, limit=1000000):
    f = open(filename, 'rb')
    chunk = f.read(limit)
    f.close()
    detected = chardet.detect(chunk)
    return detected

def detect_delimiter(filename, encoding='utf8'):
    f = open(filename, 'r', encoding=encoding)
    line = f.readline()
    f.close()
    dict1 = {',': line.count(','), ';': line.count(';'), '\t': line.count('\t'), '|' : line.count('|')}
    delimiter = max(dict1, key=dict1.get)
    return delimiter

def get_file_type(filename):
    ext = filename.rsplit('.', 1)[-1].lower()
    if ext in SUPPORTED_FILE_TYPES:
        return ext
    return None

def get_option(options, name):
    """Returns value of the option"""
    if name in options.keys():
        return options[name]
    elif name in DEFAULT_OPTIONS.keys():
        return DEFAULT_OPTIONS[name]
    return None

def write_items(fields, outdata, filetype, handle, delimiter=','):
    if len(outdata) == 0:
        return
    if filetype == 'csv':
        dw = csv.DictWriter(handle, delimiter=delimiter, fieldnames=fields)
        dw.writeheader()
        if type(outdata[0]) == type(''):
            for rawitem in outdata:
                item = {fields[0] : rawitem}
                dw.writerow(item)
        elif type(outdata[0]) == type([]):
            for rawitem in outdata:
                item = dict(zip(fields, rawitem))
                dw.writerow(item)
        else:
            dw.writerows(outdata)
    elif filetype == 'jsonl':
        # If our data is just array of strings, we just transform it to dict
        if type(outdata[0]) == type(''):
            for rawitem in outdata:
                item = {fields[0] : rawitem}
                handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
        elif type(outdata[0]) == type([]):
            for rawitem in outdata:
                item = dict(zip(fields, rawitem))
                handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
        else:
            for item in outdata:
                handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))


def get_dict_value(d, keys):
    out = []
    if d is None:
        return out
#    keys = key.split('.')
    if len(keys) == 1:
        if type(d) == type({}):
            if keys[0] in d.keys():
                out.append(d[keys[0]])
        else:
            for r in d:
                if r and keys[0] in r.keys():
                    out.append(r[keys[0]])
#        return out
    else:
        if type(d) == type({}):
            if keys[0] in d.keys():
                out.extend(get_dict_value(d[keys[0]], keys[1:]))
        else:
            for r in d:
                if keys[0] in r.keys():
                    out.extend(get_dict_value(r[keys[0]], keys[1:]))
    return out


def strip_dict_fields(record, fields, startkey=0):
    keys = record.keys()
    localf = []
    for field in fields:
        if len(field) > startkey:
            localf.append(field[startkey])
    for k in list(keys):
        if k not in localf:
            del record[k]

    if len(k) > 0:
        for k in record.keys():
            if type(record[k]) == type({}):
                record[k] = strip_dict_fields(record[k], fields, startkey + 1)
    return record


def dict_generator(indict, pre=None):
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in list(indict.items()):
            if key == "_id":
                continue
            if isinstance(value, dict):
                #                print 'dgen', value, key, pre
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    if isinstance(v, dict):
                        #                print 'dgen', value, key, pre
                        for d in dict_generator(v, pre + [key]):
                            yield d
#                    for d in dict_generator(v, [key] + pre):
#                        yield d
            else:
                yield pre + [key, value]
    else:
        yield indict



from ..utils import get_file_type, get_option
from ..constants import DATE_PATTERNS, DEFAULT_DICT_SHARE
from datetime import datetime
import logging
import orjson
import zipfile
from qddate import DateParser
from ..common.scheme import generate_scheme_from_file

class Schemer:
    def __init__(self, nodates=True):
        if nodates:
            self.qd = None
        else:
            self.qd = DateParser(generate=True)
        pass

    def generate_scheme(self, fromfile, options):
        """Generates cerberus scheme from JSON lines or BSON file"""
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        if f_type not in ['jsonl', 'bson', 'csv']:
            print('Only JSON lines, CSV and BSON (.jsonl, .csv, .bson) files supported now')
            return
        if options['zipfile']:
            z = zipfile.ZipFile(fromfile, mode='r')
            fnames = z.namelist()
            finfilename = fnames[0]
            if f_type == 'bson':
                infile = z.open(fnames[0], 'rb')
            else:
                infile = z.open(fnames[0], 'r')
        else:
            finfilename = fromfile
            if f_type == 'bson':
                infile = open(fromfile, 'rb')
            else:
                infile = open(fromfile, 'r', encoding=get_option(options, 'encoding'))

        logging.debug('Start identifying scheme for %s' % (fromfile))
        scheme = generate_scheme_from_file(fileobj=infile, filetype=f_type, delimiter=options['delimiter'], encoding=options['encoding'])
        if options['output']:
            f = open(options['output'], 'wb', encoding='utf8')
            f.write(orjson.dumps(scheme, option=orjson.OPT_INDENT_2))
            f.close()
        else:
            print(str(orjson.dumps(scheme, option=orjson.OPT_INDENT_2)))
